"""
Bridging code between the MNIST data format and the vector representation used
in the network class.
"""

import mnist
import numpy as np

def vectorize_image(image):
    """
    Maps an MNIST image to a vector (more specifically, a "numpy.ndarray"). An
    MNIST image is a 784-component tuple of integers between 0 and 255,
    representing greyscale values. The resulting vector is a 784-dimensional
    column vector. Each greyscale value is normalized to the interval "[0, 1]".
    """
    x = np.reshape(np.array(image, dtype=np.float32), (784, ))
    x *= 1.0 / 255
    return x

def vectorize_label(label):
    """
    Maps an MNIST label to a vector (more specifically, a "numpy.ndarray"). An
    MNIST label is an integer "i" between 0 and 9. The resulting vector is a
    10-dimensional column vector with a 1.0 in the "i"th position and zeroes
    elsewhere.
    """
    #x = np.zeros((10, 1), dtype=np.float32)
    #x[label] = 1.0
    
    x = np.array(label)
    return x

def read_training_data():
    """
    Reads training data. This returns a list of pairs "(x, y)" where "x" is a
    784-dimensional vector (more specifically, a "numpy.ndarray") representing
    the image and "y" is a 10-dimensional one-hot vector representing the label.
    """
    data = mnist.read_training_data()
    return [(vectorize_image(i_l[0]), vectorize_label(i_l[1])) for i_l in data]

def read_test_data():
    """
    Reads test data. This returns a list of pairs "(x, y)" where "x" is a
    784-dimensional vector (more specifically, a "numpy.ndarray") representing
    the image and "y" is an integer representing the label.
    """
    data = mnist.read_test_data()
    return [(vectorize_image(i_l1[0]), i_l1[1]) for i_l1 in data]
