"""
Feed-forward neural network in PyTorch. Adapted to do training/testing on the version of the
MNIST-dataset used in earlier versions of this lab.

The code is heavily based on:
https://github.com/pytorch/examples/blob/main/mnist/main.py

"""

import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

import mnist_network
import mnist_dataset
import network_config


class Net(nn.Module):
    """
    The neural network class. This is a feed-forward network with ReLU activation and final softmax.
    """

    def __init__(self, network_conf):
        """
        Create the network according to the given configuration
        args:
            network_conf: The configuration of the network represented as a list of tuples with the number of nodes in each layer
        """
        super(Net, self).__init__()

        self.layers = self.create_network_modules(network_conf=network_conf)

    def forward(self, x):
        """
        Forward pass through the network
        """
        # Iterate over the layers and pass the input through them
        for layer in self.layers:
            x = layer(x)

        return x

    def create_network_modules(self, network_conf):
        """
        Create the network structure according to the given configuration
        args:
            network_conf: The configuration of the network represented as a list of tuples with the number of nodes in each layer
        returns:
            A list of network modules
        """
        # Create a list of modules
        module_list = nn.ModuleList()
        # Iterate over the configuration
        for i, layer in enumerate(network_conf):
            # Add a linear layer
            module_list.append(nn.Linear(layer[0], layer[1]))

            if i != len(network_conf) - 1:
                # Add the ReLU activation function after each layer except the last one
                module_list.append(nn.ReLU())
            else:
                # on the last layer, add the softmax instead
                module_list.append(nn.LogSoftmax(dim=1))

        print(f"Initialized a neural network with the following configuration: \n{module_list}")
        return module_list


def train(args, model, device, train_loader, optimizer, epoch):
    """
    Train the network

    Args:
        args: The command line arguments
        model: The network model
        device: The device to run the training on
        train_loader: The training data loader
        optimizer: The optimizer to use
        epoch: Current epoch
    """
    # Set the model to training mode
    model.train()
    # Iterate over the training data in batches
    for batch_idx, (data, target) in enumerate(train_loader):
        # Move the data to the device
        data, target = data.to(device), target.to(device)
        # Zero out the gradients
        optimizer.zero_grad()
        # Forward pass
        output = model(data)
        # Calculate the loss
        loss = F.nll_loss(output, target)
        # Backward pass
        loss.backward()
        # Update the weights
        optimizer.step()
        # Print the progress at the specified intervals
        if batch_idx % args.log_interval == 0:
            print(
                "Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}".format(
                    epoch,
                    batch_idx * len(data),
                    len(train_loader.dataset),
                    100.0 * batch_idx / len(train_loader),
                    loss.item(),
                )
            )
            # Check if we should do a dry run for debugging
            if args.test_run:
                print("Test run, exiting after one batch")
                print("Target shape: ", target.shape)
                print("Output shape: ", output.shape)
                print("Test succeeded, exiting")
                break
                


def test(model, device, test_loader):
    """
    Test the network
    Args:
        model: The network model
        device: The device to run the testing on
        test_loader: The test data loader
    """
    # Set the model to evaluation mode
    model.eval()
    # Initialize the loss and number of correct predictions
    test_loss = 0
    correct = 0
    # Disable gradient calculation
    with torch.no_grad():
        # Iterate over the test data
        for data, target in test_loader:
            # Move the data to the device
            data, target = data.to(device), target.to(device)

            # Forward pass
            output = model(data)
            # sum up batch loss
            test_loss += F.nll_loss(output, target, reduction="sum").item()
            # get the index of the max log-probability
            pred = output.argmax(dim=1, keepdim=True)
            # Add the number of correct predictions
            # elementwise comparison between the prediction and the target after converting the target to the same shape as the prediction
            correct += pred.eq(target.view_as(pred)).sum().item()
    # Calculate the average loss
    test_loss /= len(test_loader.dataset)
    # Print the results
    print(
        "\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n".format(
            test_loss,
            correct,
            len(test_loader.dataset),
            100.0 * correct / len(test_loader.dataset),
        )
    )


def main():
    # Training settings
    parser = argparse.ArgumentParser(description="PyTorch MNIST Example")
    parser.add_argument(
        "--batch-size",
        type=int,
        default=16,
        metavar="N",
        help="input batch size for training (default: 16)",
    )
    parser.add_argument(
        "--test-batch-size",
        type=int,
        default=64,
        metavar="N",
        help="input batch size for testing (default: 64)",
    )
    parser.add_argument(
        "--hidden-size",
        type=int,
        default=20,
        metavar="H",
        help="Size of the hidden layer (default: 20)",
    )
    parser.add_argument(
        "--epochs",
        type=int,
        default=1,
        metavar="N",
        help="Number of epochs to train (default: 1)",
    )
    parser.add_argument(
        "--lr",
        type=float,
        default=1.0,
        metavar="LR",
        help="Learning rate (default: 1.0)",
    )
    parser.add_argument(
        "--test-run",
        action="store_true",
        default=False,
        help="Quickly test a single pass (used for debugging)",
    )
    parser.add_argument(
        "--log-interval",
        type=int,
        default=500,
        metavar="N",
        help="How many batches to wait before logging training status",
    )
    parser.add_argument(
        "--run-vg",
        action="store_true",
        default=False,
        help="Run the code with the VG-config",
    )

    args = parser.parse_args()

    run_vg = args.run_vg
    if run_vg:
        # The VG version runs the config specified by the user
        network_conf = network_config.vg_config
    else:
        # The G version is using a config of three layers:
        # input(784 nodes)
        # hidden(hidden_size nodes)
        # output(10 nodes)
        hidden_size = args.hidden_size
        network_conf = [(784, hidden_size), (hidden_size, 10)]

    # Set the random seed
    torch.manual_seed(42)

    # Set the device to run the training on, for this lab we will only use the CPU
    device = torch.device("cpu")

    # Set the kwargs for the data loaders
    train_kwargs = {"batch_size": args.batch_size}
    test_kwargs = {"batch_size": args.test_batch_size}

    # Load the data
    print("Loading the training data")
    training_data = mnist_network.read_training_data()
    print("Loading the test data")
    test_data = mnist_network.read_test_data()

    # Wrap the data in a custom dataset
    training_data = mnist_dataset.CustomMnistDataset(training_data)
    test_data = mnist_dataset.CustomMnistDataset(test_data)

    # Create the data loaders
    train_loader = torch.utils.data.DataLoader(training_data, **train_kwargs)
    test_loader = torch.utils.data.DataLoader(test_data, **test_kwargs)
    # Create the network and optimizer
    model = Net(network_conf).to(device)
    optimizer = optim.SGD(model.parameters(), lr=args.lr)

    # Train the network for the specified number of epochs
    for epoch in range(1, args.epochs + 1):
        train(args, model, device, train_loader, optimizer, epoch)
    # Test the network
    if not args.test_run:
        test(model, device, test_loader)


if __name__ == "__main__":
    main()
